/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2002 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */


#include <stdio.h>
#include <malloc.h>

#include "map.h"

#define MAP_SIZE 5000

void *safe_malloc (size_t size);
void *safe_realloc (void *current, size_t size);


struct element
{
  void *key;
  void *value;

  struct element *next;
};

struct map
{
  int (*compare) (void *, void *);
  struct element *buckets[MAP_SIZE];
  size_t size_of_key;
  size_t size_of_value;
};

/* An hash function based on the address:*/
int
hash_function (void *x)
{
  int r = (((int) *(void **) x) / 8) % MAP_SIZE;
  //printf("{%i}",r);
  return r;
}

void
destroy_bucket (struct element *b)
{
  struct element *next;
  struct element *current = b;

  while (current != NULL)
    {
      next = current->next;
      free (current->key);
      free (current->value);
      free (current);
      current = next;
    }				/* while */
}

struct map *
create_map (int (*compare_) (void *, void *),
	    size_t size_of_key_, size_t size_of_value_)
{
  int i;
  struct map *r = (struct map *) safe_malloc (sizeof (struct map));

  r->compare = compare_;
  r->size_of_key = size_of_key_;
  r->size_of_value = size_of_value_;

  for (i = 0; i < MAP_SIZE; i++)
    r->buckets[i] = NULL;

  //  printf("<C>");
  return r;
}

void
destroy_map (struct map *m)
{
  int i;
  int fb = 0;
  for (i = 0; i < MAP_SIZE; i++)
    {
      if (m->buckets[i] != NULL)
	fb++;
      destroy_bucket (m->buckets[i]);
    }

  //  printf("<D %f>",(float)fb/(float)MAP_SIZE);

  free (m);
}

struct element *
search_in_bucket (struct map *m, struct element *bucket, void *k)
{
  struct element *x = bucket;

  while (x != NULL)
    {
      if (!m->compare (x->key, k))
	return x;		/* We have found k */
      else
	x = x->next;
    }

  /* x is NULL and we have not found k: */
  return NULL;
}

void
insert_into_map (struct map *m, void *k, void *v)
{
  int bucketNo = hash_function (k);

  if (search_in_bucket (m, m->buckets[bucketNo], k) == NULL)
    {
      struct element *e =
	(struct element *) safe_malloc (sizeof (struct element));

      e->key = safe_malloc (m->size_of_key);
      memcpy (e->key, k, m->size_of_key);

      e->value = safe_malloc (m->size_of_value);
      memcpy (e->value, v, m->size_of_value);

      e->next = m->buckets[bucketNo];
      m->buckets[bucketNo] = e;
    }
}

void *
access_map (struct map *m, void *k)
{
  int bucketNo = hash_function (k);
  struct element *e;

  if ((e = search_in_bucket (m, m->buckets[bucketNo], k)) == NULL)
    return NULL;
  else
    return e->value;
}

int
int_compare (void *a, void *b)
{
  return *(int *) b - *(int *) a;
}

int
double_compare (void *a, void *b)
{
  if (*(double *) a < *(double *) b)
    return -1;
  else if (*(double *) a == *(double *) b)
    return 0;
  else
    return +1;
}

int
string_compare (void *a, void *b)
{
  return strcmp ((char *) a, (char *) b);
}

int
pointer_compare (void *a, void *b)
{
  if (*(void **) a == *(void **) b)
    return 0;
  else if (*(void **) a < *(void **) b)
    return -1;
  else
    return +1;
}
